#!/usr/bin/php
<?php

use Database\Driver\DefaultDB;

require_once("/usr/local/opmon/etc/config.php");

define(FILE, '/var/log/opmon-set-permissions.log');

/**
 * Connection with OpMon database.
 * @return DefaultDB
 */
function getConnection()
{
    return new DefaultDB();
}

/**
 * Print in StdIN and log in /var/log/messages
 * @param string $msg
 */
function logger($msg)
{
    global $options;

    print $msg . PHP_EOL;

    if (isset($options['force'])) {
        Logger::log("Set-Permission", $msg);
    }
}

/**
 * Execute query in database.
 * @param string $query
 * @param array $params
 * @return int|mixed
 */
function doQuery($query, $params = null)
{
    $db = getConnection();
    $db->prepare($query);
    $db->execute($params);

    if (preg_match("/DELETE/", $query)) {
        return true;
    }
    return $db->fetchAll($db::FETCH_ASSOC);
}

/**
 * Verify if user accepts continue.
 * @param string $msg
 * @return bool
 */
function canContinue($msg = "All permissions are removed, do you want to continue? [y/n]: ")
{
    $line = strtolower(readline($msg));
    switch ($line) {
        case "y":
            return true;
        case "n":
            return false;
        default:
            return canContinue($msg);
    }
}

/**
 * Make a backup of opcfg database.
 * @throws Exception
 */
function makeBackupOpCfg()
{
    $pluginBackup = "/usr/local/opmon/utils/opmon-base.pl";
    if (!file_exists($pluginBackup) || !is_executable($pluginBackup)) {
        throw new Exception("File $pluginBackup doesn't exist or is not executable");
    }
    $cmd = $pluginBackup . " -e opcfg 2&>1 >/dev/null";
    system($cmd);
}

/**
 * Rename folder backup of opcfg to stay with plugin execution date.
 * @return string
 * @throws Exception
 */
function renameFolderBackup()
{
    $now = date('dmyHis');
    $folderBackupBefore = "/var/tmp/opmondb/opcfg";
    $folderBackupAfter = "$folderBackupBefore-$now";

    if (!rename($folderBackupBefore, $folderBackupAfter)) {
        throw new Exception("Error to rename folder backup");
    }
    return $folderBackupAfter;
}

/**
 * Remove registers of contacts tables in opcfg
 */
function removePermissions()
{
    canContinue() ?: exit(0);
    makeBackupOpCfg();
    $newFolderBackup = renameFolderBackup();
    $deletion = array(
        'DELETE FROM `opcfg`.`nagios_host_contactgroups`',
        'DELETE FROM `opcfg`.`nagios_host_contacts`',
        'DELETE FROM `opcfg`.`nagios_host_template_contactgroups`',
        'DELETE FROM `opcfg`.`nagios_host_template_contacts`',
        'DELETE FROM `opcfg`.`nagios_service_contactgroups`',
        'DELETE FROM `opcfg`.`nagios_service_contacts`',
        'DELETE FROM `opcfg`.`nagios_service_template_contactgroups`',
        'DELETE FROM `opcfg`.`nagios_service_template_contacts`'
    );

    foreach ($deletion as $del) {
        doQuery($del);
    }

    logger("The permissions were removed and the backup was moved to $newFolderBackup");
    exit(0);
}

/**
 * Show help to use plugin.
 */
function help()
{
    $msg = "
    This plugin aims to config permissions in OpMon through the following options:
        - Del contactgroups and contacts of all hosts and OpMon services.
        - Add contactgroups in services of servicegroup;
        - Add contactgroups in hosts of hostgroup;
        - Add services in servicegroup;
        - Add hosts in hostgroup;

    General options:
    --remove    Remove all contacts and contactgroups of hosts and OpMon services, making a backup before from opcfg
                database in /var/tmp/opmondb.
    --option    Options to run the plugin:
                    --addhosttohostgroup                Add host in hostgroup
                    --addservicetoservicegroup          Add service in servicegroup
                    --addcontactgrouptoservices         Add contactgroup in all services in servicegroup
                    --addcontactgrouptohosts            Add contactgroup in all hosts in hostgroup
    --service       Service to add in servicegroup
    --host          Host to add in hostgroup or filter
    --contactgroup  Contactgroup to add and servicegroups or hostgroups
    --servicegroup  Servicegroup that will receive the services
    --hostgroup     Hostgroup that will receive the hosts
    --testrun       It shows what will be applied of additional in servicegroups and hostgroups and log in
                    /var/log/opmon-set-permissions.log, this parameter is not valid to addcontactgrouptoservices and
                    addcontactgrouptohosts.
    --regex     Use regular expression to filter hosts and services
    --force     Log configurations applied in log /var/log/messages

    Ex.:
    The command bellow will remove all contacts of hosts and services:
        ./opmon-set-permission.php --option=remove

    The command bellow will add in servicegroup all services that contain the word 'disk' and the hosts
    that start with the word 'linux':
        ./opmon-set-permission.php --option=addservicetoservicegroup --host='/^linux/i' --service='/disk/i' --servicegroup disks --regex

    The command bellow will add in servicegroup 'cpu' the service 'CPU' of host 'opmon':
        ./opmon-set-permission.php --option=addservicetoservicegroup --host='opmon' --service='CPU' --servicegroup='cpu'

    The command bellow will verify if the host 'windows-server' will be added in hostgroup 'windows' and will log in
    /var/log/opmon-set-permissions.log:
        ./opmon-set-permission.php --option=addhosttohostgroup --host='windows-server' --hostgroup='windows' --testrun

    The command bellow will add in hostgroup 'windows' all hosts that begin with the word 'windows' and the result will be logged
        in /var/log/messages:
        ./opmon-set-permission.php --option=addhosttohostgroup --host='/^windows/' --hostgroup='windows' --regex --force

    The command bellow will add the contactgroup 'operacao' in all services of servicegroup 'gop':
        ./opmon-set-permission.php --option=addcontactgrouptoservices --contactgroup='operacao' --servicegroup='gop'

    The command bellow will add the contactgroup 'opmonadmins' in all hosts of hostgroup 'servers':
    ./opmon-set-permission.php --option=addcontactgrouptohosts --contactgroup='opmonadmins' --hostgroup='servers'
    ";
    print $msg . PHP_EOL;
}

/**
 * Get parameters provided to plugin.
 * @return array
 */
function getOptions()
{
    $longOptions = array(
        "regex",
        "help",
        "host:",
        "hostgroup:",
        "service:",
        "servicegroup:",
        "contactgroup:",
        "force",
        "testrun",
        "option:"
    );

    $options = getopt('h', $longOptions);

    if (isset($options['h']) || isset($options['help'])) {
        help();
        exit(0);
    }
    return $options;
}

/**
 * Validate parameters.
 * @throws Exception
 * @throws InvalidArgumentException
 */
function validate()
{
    global $options;

    if (isset($options['testrun']) && isset($options['force'])) {
        throw new Exception("Parameter --force and --forcerun can't use together");
    }

    switch ($options['option']) {
        case "remove":
            return;
        case "addservicetoservicegroup":
            if (is_null($options['service']) || is_null($options['servicegroup']) || is_null($options['host'])) {
                throw new Exception("Arguments missing");
            }
            return;
        case "addhosttohostgroup":
            if (is_null($options['hostgroup']) || is_null($options['host'])) {
                throw new Exception("Arguments missing");
            }
            return;
        case "addcontactgrouptohosts":
            if (is_null($options['hostgroup']) || is_null($options['contactgroup'])) {
                throw new Exception("Arguments missing");
            }
            if (isset($options['testrun'])) {
                throw new Exception("Testrun not enable to this option");
            }
            return;
        case "addcontactgrouptoservices":
            if (is_null($options['servicegroup']) || is_null($options['contactgroup'])) {
                throw new Exception("Arguments missing");
            }
            if (isset($options['testrun'])) {
                throw new Exception("Testrun not enable to this option");
            }
            return;
        default:
            throw new InvalidArgumentException("Argument invalid");
    }

}

/**
 * Get the id of contactgroup, servicegroup or hostgroup.
 * @param ConfigManager $manager
 * @param string $name
 * @param string $object
 * @return integer $groupId
 * @throws Exception
 */
function getGroupIdByName($manager, $name, $object = "host")
{
    if ($object == "host") {
        $groupId = $manager->cfg->return_hostgroup_id_by_name($name);
        $errorMsg = "Host group [ $name ] doesn't exist";
    } elseif ($object == "service") {
        $groupId = $manager->cfg->return_servicegroup_id_by_name($name);
        $errorMsg = "Service group [ $name ] doesn't exist";
    } elseif ($object == "contactgroup") {
        $groupId = $manager->cfg->return_contactgroup_id_by_name($name);
        $errorMsg = "Contact group [ $name ] doesn't exist";
    }

    if (!$groupId) {
        throw new Exception($errorMsg);
    }
    return $groupId;
}

/**
 * Logger the elements that will be added
 * @param array $objects
 * @param string $objectType
 */
function testRun($objects, $objectType = 'host')
{
    global $options;
    $group = $options['hostgroup'] ?: $options['servicegroup'];
    $total = count($objects);

    $msg = "Will be added [ $total ] $objectType(s) in {$objectType}group [ $group ], to more detail view " . FILE;

    if ($total == 0) {
        $msg = "No $objectType will be added";
        print $msg;
        exit(0);
    }

    print $msg;
    foreach ($objects as $object) {
        if ($objectType == 'host') {
            $msg = sprintf(
                "Host [ %s ] will be added in %sgroup [ %s ]" . PHP_EOL,
                $object['host_name'],
                $objectType,
                $group
            );
        } else {
            $msg = sprintf(
                "Service [ %s ] of host [ %s ] will be added in %sgroup [ %s ]" . PHP_EOL,
                $object['service_description'],
                $object['host_name'],
                $objectType,
                $group
            );
        }
        file_put_contents(FILE, date("d/m/Y - G:i:s -> ") . $msg, FILE_APPEND);
    }
}

/**
 * Add each host in hostgroup
 * @param ConfigManager $manager
 */
function addHostToHostGroup($manager)
{
    global $options;
    $hostGroupId = getGroupIdByName($manager, $options['hostgroup']);
    $hosts = getHosts($manager, $options['host'], $hostGroupId, isset($options['regex']));

    if (isset($options['testrun'])) {
        testrun($hosts);
        exit(0);
    }

    if (count($hosts) == 0) {
        $msg = "No host was added";
        logger($msg);
    }

    foreach ($hosts as $host) {
        $manager->cfg->add_hostgroup_member($hostGroupId, $host['host_id']);
        $msg = "Host [ {$host['host_name']} ] was added to hostgroup [ {$options['hostgroup']} ]";
        logger($msg);
    }
}

/**
 * Add each service in servicegroup
 * @param ConfigManager $manager
 */
function addServiceToServiceGroup($manager)
{
    global $options;
    $serviceGroupId = getGroupIdByName($manager, $options['servicegroup'], 'service');
    $services = getServices($manager, $options['service'], $serviceGroupId, $options['host'], isset($options['regex']));

    if (isset($options['testrun'])) {
        testrun($services, 'service');
        exit(0);
    }

    if (count($services) == 0) {
        $msg = "No service was added";
        logger($msg);
    }

    foreach ($services as $service) {
        $manager->cfg->add_service_servicegroup($service['service_id'], $serviceGroupId);
        $msg = sprintf(
            "Service [ %s ] of host [ %s ] was added to servicegroup [ %s ]",
            $service['service_description'],
            $service['host_name'],
            $options['servicegroup']
        );
        logger($msg);
    }
}

/**
 * Filter only services that must be added in servicegroup.
 * @param ConfigManager $manager
 * @param string $serviceToAdd
 * @param integer $serviceGroupId
 * @param string $filterHost
 * @param bool $useRegex
 * @return array|int|mixed
 * @throws Exception
 */
function getServices($manager, $serviceToAdd, $serviceGroupId, $filterHost, $useRegex = false)
{
    $servicesToReturn = array();

    if (!$useRegex) {
        $query = 'SELECT `nh`.`host_name`, `ns`.`service_description`, `ns`.`service_id`
                  FROM `opcfg`.`nagios_hosts` AS `nh` 
                  JOIN `opcfg`.`nagios_services` AS `ns` ON `ns`.`host_id` = `nh`.`host_id`
                  WHERE `nh`.`host_name` = :host AND `ns`.`service_description` = :service';
        $service = doQuery($query, array(':host' => $filterHost, ':service' => $serviceToAdd));

        if (!$service[0]['service_id']) {
            throw new Exception("Service [ $serviceToAdd ] doesn't exist in host [ $filterHost ]");
        }

        $service = ($manager->cfg->servicegroup_has_member($serviceGroupId, $service[0]['service_id']) ? array() : $service);
        return $service;
    }
    $allOpMonServices = array();
    $manager->cfg->get_services_list($allOpMonServices);

    $services = array_filter($allOpMonServices, function ($service) {
        return isset($service['host_id']);
    });

    $allOpMonHosts = array();
    $manager->cfg->get_host_info_list($allOpMonHosts);
    $allOpMonHosts = array_filter($allOpMonHosts, function ($host) use ($filterHost) {
        preg_match($filterHost, $host['host_name'], $match);
        return $match ? 1 : 0;
    });

    foreach ($services as $service) {
        preg_match($serviceToAdd, $service['service_description'], $match);
        if ($match && !$manager->cfg->servicegroup_has_member($serviceGroupId, $service['service_id'])) {
            foreach ($allOpMonHosts as $host) {
                if ($service['host_id'] == $host['host_id']) {
                    $service['host_name'] = $host['host_name'];
                    $servicesToReturn[] = $service;
                }
            }
        }
    }
    return $servicesToReturn;
}

/**
 * Filter only hosts that must be added in hostgroup.
 * @param ConfigManager $manager
 * @param string $hostToAdd
 * @param integer $hostGroupId
 * @param bool $useRegex
 * @return array|int|mixed
 * @throws Exception
 */
function getHosts($manager, $hostToAdd, $hostGroupId, $useRegex = false)
{
    if (!$useRegex) {
        $query = 'SELECT host_id, host_name FROM `opcfg`.`nagios_hosts` WHERE `host_name` = ?';
        $host = doQuery($query, array($hostToAdd));

        if (!$host[0]['host_id']) {
            throw new Exception("Host [ $hostToAdd ] doesn't exist");
        }

        $host = ($manager->cfg->hostgroup_has_member($hostGroupId, $host[0]['host_id']) ? array() : $host);
        return $host;
    }

    $allOpMonHosts = array();
    $manager->cfg->get_host_info_list($allOpMonHosts);
    $hosts = array();

    foreach ($allOpMonHosts as $host) {
        preg_match($hostToAdd, $host['host_name'], $match);
        if ($match && !$manager->cfg->hostgroup_has_member($hostGroupId, $host['host_id'])) {
            $hosts[] = $host;
        }
    }
    return $hosts;
}

/**
 * Add contactgroup in each host of a hostgroup.
 * @param ConfigManager $manager
 */
function addContactGroupToHosts($manager)
{
    global $options;
    $contactGroupId = getGroupIdByName($manager, $options['contactgroup'], "contactgroup");
    $hostGroupId = getGroupIdByName($manager, $options['hostgroup']);
    $hosts = array();
    $manager->cfg->return_hostgroup_member_list($hostGroupId, $hosts);

    foreach ($hosts as $host) {
        if (!$manager->cfg->host_has_contactgroup($host['host_id'], $contactGroupId)) {
            $manager->cfg->add_host_contactgroup($host['host_id'], $contactGroupId);
        }
    }
    $msg = sprintf(
        "Contact group [ %s ] was added to hosts in hostgroup  [ %s ]",
        $options['contactgroup'],
        $options['hostgroup']
    );
    logger($msg);
}

/**
 * Add contactgroup in each service of a servicegroup.
 * @param ConfigManager $manager
 */
function addContactGroupToServices($manager)
{
    global $options;
    $contactGroupId = getGroupIdByName($manager, $options['contactgroup'], "contactgroup");
    $serviceGroupId = getGroupIdByName($manager, $options['servicegroup'], 'service');
    $services = array();
    $manager->cfg->return_servicegroup_member_list($serviceGroupId, $services);
    foreach ($services as $service) {
        if (!$manager->cfg->service_has_contactgroup($service['service_id'], $contactGroupId)) {
            $manager->cfg->add_service_contactgroup($service['service_id'], $contactGroupId);
        }
    }
    $msg = sprintf(
        "Contact group [ %s ] was added to services in servicegroup  [ %s ]",
        $options['contactgroup'],
        $options['servicegroup']
    );
    logger($msg);
}

/**
 * Main plugin routine.
 */
function main()
{
    try {
        global $options;
        $options = getOptions();
        validate();
        $manager = new ConfigManager();
        switch ($options['option']) {
            case "remove":
                return removePermissions();
            case "addcontactgrouptohosts":
                return addContactGroupToHosts($manager);
            case "addcontactgrouptoservices":
                return addContactGroupToServices($manager);
            case "addhosttohostgroup":
                return addHostToHostGroup($manager);
            case "addservicetoservicegroup":
                return addServiceToServiceGroup($manager);
        }
    } catch (Exception $ex) {
        logger($ex->getMessage());
    }
}
main();