#!/usr/bin/php
<?php

include("/usr/local/opmon/etc/config.php");

use Database\Driver\DefaultDB as DB;
set_logged_user("opmonadmin");

$config = new ConfigManager();

function logger($msg)
{
    $msg = str_replace("\n", "", $msg);

    $date = new DateTime();
    $dateFormated = $date->format('Y-m-d H:i:s');

    $logFile = fopen("/var/log/opmon-topology.log", "a");

    $text = "[$dateFormated] $msg".PHP_EOL;
    fprintf($logFile, $text);
    fclose($logFile);
    print $text;
}

function runTraceroute($address)
{
    $traceroute = new TraceRouteIntegration();
    $traceroute->set_host_ip($address);
    //$traceroute->set_debug_on();
    return array_reverse($traceroute->run());
}

function getPropertie($node, $propertie)
{
    foreach ($node->attributes->entry as $properties) {
        if ($properties->key === $propertie) {
            return $properties->value;
        }
    }
}

function getHostIdByAddress($address)
{
    $db = new DB();
    $db->prepare(
        'SELECT host_id from opcfg.nagios_hosts where address = :address'
    );
    $db->execute(array(':address' => $address));
    return $db->fetch(DB::FETCH_OBJ)->host_id;
}

function ping($address)
{
    exec(
        "/bin/ping -n -w 1 -c 1 {$address}",
        $output,
        $code
    );
    return ($code == 0);
}

function getParents($host_id)
{
    $db = new DB();

    $db->prepare('
        SELECT parent_id FROM (
                SELECT parents AS parent_id
                FROM opcfg.nagios_hosts
                WHERE host_id = :host_id
            UNION
                SELECT parent_id
                FROM opcfg.nagios_host_parents
                WHERE child_id = :host_id ) AS t
        WHERE parent_id IS NOT NULL
        ');

    $db->execute(
        array(
            ':host_id' => $host_id
        )
    );

    return $db->fetchAll(DB::FETCH_COLUMN);
}

function getHostAlreadyMonitoredFromAddress($address)
{
    global $config;

    $hopId = null;
    if ($config->cfg->host_address_exists($address)) {
        $hopId = getHostIdByAddress($address);
    }

    return (is_numeric($hopId)) ? $hopId : null;
}

function main()
{
    global $config;

    logger("Starting topology");

    /* get opmon host id */
    $opmon_id = $config->cfg->return_host_id_by_name("opmon");
    if ($opmon_id == 0) {
         $opmon_id = $config->addHost(
            array(
                "host_name" => "opmon",
                "alias" => "OpMon host from where the topology starts",
                "address" => '127.0.0.1'
            )
        );
        $config->associateContactGroupsToHost($opmon_id, array(1));

    }

    $hosts = array();
    $config->cfg->get_host_list($hosts);

    $count = 0;
    foreach ($hosts as $host) {
        $count++;
        logger("Mapping $count of ".count($hosts));
        if (isset($host['parents']) && $host['parents'] > 0) {
            print $host['host_name']." has parent ".$host['parents']."\n";
            continue;
        }
        if ($host['host_name'] == 'opmon') {
            continue;
        }
        print $host['host_name']." ".$host['address']."\n";

        $id = $host['host_id'];
        $name = $host['host_name'];
        $address = $host['address'];
        $route = runTraceroute($address);
        if (empty($route)) {
            $config->modifyHost(
                array(
                    'host_id' => $id,
                    'parents' => $opmon_id
                )
            );
            $modified = true;
            logger("Set parent child {$id} $opmon_id");
            continue;
        }

        logger(PHP_EOL);
        logger("Run traceroute to host {$name} id({$id}) address({$address})");
        foreach ($route as $hop) {
            if (ping($hop['ip']) === false) {
                continue;
            }

            $hopId = getHostAlreadyMonitoredFromAddress($hop['ip']);
            if (is_null($hopId)) {
                $hopId = $config->cfg->return_host_id_by_name("Hop {$hop['name']}");
                if ($hopId == 0) {
                    $hopId = null;
                }
                logger("Found by name Hop {$hop['name']} {$hopId}");
            }

            if (is_null($hopId)) {
                logger("Create a new host: Hop {$hop['name']}");

                 $hopId = $config->addHost(
                    array(
                        "host_name" => "Hop {$hop['name']}",
                        "alias" => "Dynamic Topology Hop Interface IP",
                        "address" => $hop['ip']
                    )
                );
                $config->associateContactGroupsToHost($hopId, array(1));
            } else {
                logger("Loading host id({$hopId}) with {$hop['ip']}");
            }

            $childParents = getParents($id);
            if (! empty($childParents)) {
                if (in_array($hopId, $childParents)) {
                    logger("Skiping child {$id} ip({$address}) already father {$hopId} ip({$hop['ip']})");
                    $address = $hop['ip'];
                    $id = $hopId;
                    continue;
                }
                $config->cfg->host_add_parent($id, $hopId);
                $modified = true;
                logger("Set aditional parent child {$id} ip({$address}) > father {$hopId} ip({$hop['ip']})");
                $address = $hop['ip'];
                $id = $hopId;
            } else {
                $config->modifyHost(
                    array(
                        'host_id' => $id,
                        'parents' => $hopId
                    )
                );
                $modified = true;
                logger("Set parent child {$id} ip({$address}) > father {$hopId} ip({$hop['ip']})");
                $address = $hop['ip'];
                $id = $hopId;
            }
        }
    }

    if ($modified === true) {
        $config->commit(1, true);
        logger("Finish applied changes");
        exit;
    }

    logger("Finish without changes");
    exit;
}
main();
